from typing import Dict, List, Literal, Optional, Tuple

import httpx
import openai
from nonebot import logger
from openai import AsyncOpenAI
from pydantic import BaseModel

from nonebot_plugin_dice_narrator.config import config

openai.base_url = config.OPENAI_BASE_URL
openai_key_idx = 0


class RunOutOfKey(Exception):
    pass


def _key_iterator_decorator(openai_req_func):
    async def wrapper(**kwargs):
        global openai_key_idx

        try_cnt = len(config.OPENAI_API_KEYS)
        while try_cnt > 0:
            try_cnt -= 1
            openai_key_idx = (openai_key_idx + 1) % len(config.OPENAI_API_KEYS)
            openai.api_key = config.OPENAI_API_KEYS[openai_key_idx]
            try:
                return await openai_req_func(
                    api_key=(
                        config.OPENAI_API_KEYS[openai_key_idx]
                        if "api_key" not in kwargs
                        else kwargs["api_key"]
                    ),
                    **kwargs,
                )
            except Exception as e:
                logger.error(
                    f"Failed to run {openai_req_func.__name__} with Exception: {e}",
                )
        else:
            raise RunOutOfKey

    return wrapper


@_key_iterator_decorator
async def gen_chat_response_text(
    messages: List[Dict],
    temperature: float = 0,
    frequency_penalty: float = 0.2,
    presence_penalty: float = 0.2,
    top_p=1,
    api_key: Optional[str] = None,
) -> Tuple[str, int]:
    """生成聊天回复内容"""

    client = AsyncOpenAI(
        api_key=api_key,
        base_url=config.OPENAI_BASE_URL,
        http_client=(
            httpx.AsyncClient(
                proxies=config.OPENAI_PROXY,
            )
            if config.OPENAI_PROXY
            else None
        ),
    )

    res = await client.chat.completions.create(
        model=config.CHAT_MODEL,
        messages=messages,  # type: ignore
        temperature=temperature,
        frequency_penalty=frequency_penalty,
        presence_penalty=presence_penalty,
        top_p=top_p,
    )
    logger.info(res)

    output = res.choices[0].message.content
    token_consumption = res.usage.total_tokens if res.usage else -1
    logger.info(f"请求消耗: {token_consumption} tokens")
    assert output, "Chat response is empty"
    return output, token_consumption



from pathlib import Path
from typing import List, Optional

import miose_toolkit_common.config
from miose_toolkit_common.config import Config, Env

miose_toolkit_common.config._config_root = (  # noqa: SLF001
    Path("./configs/nonebot_plugin_dice_narrator")
)


class PluginConfig(Config):
    OPENAI_API_KEYS: List[str] = []
    OPENAI_BASE_URL: Optional[str] = None
    OPENAI_PROXY: Optional[str] = None
    CHAT_MODEL: str = "gpt-3.5-turbo"

    DIFFICULTY_DICE_MAX: int = 20

    FORBIDDEN_USERS: List[str] = []


config = PluginConfig().load_config(create_if_not_exists=True)
config.dump_config(envs=[Env.Default.value])



from typing import Optional, Tuple, Union

from nonebot.adapters import Bot
from nonebot.adapters.onebot.v11 import (
    GroupIncreaseNoticeEvent,
    GroupMessageEvent,
    MessageEvent,
)


async def gen_chat_text(event: MessageEvent, bot: Bot) -> Tuple[str, bool]:
    """生成合适的会话消息内容(eg. 将cq at 解析为真实的名字)"""
    if not isinstance(event, GroupMessageEvent):
        return event.get_plaintext(), False

    wake_up = False
    msg = ""
    for seg in event.message:
        if seg.is_text():
            msg += seg.data.get("text", "")
        elif seg.type == "at":
            qq = seg.data.get("qq", None)
            if qq:
                if qq == "all":
                    msg += "@全体成员"
                    wake_up = True
                else:
                    user_name = await get_user_name(
                        event=event,
                        bot=bot,
                        user_id=int(qq),
                    )
                    if user_name:
                        msg += (
                            f"@{user_name}"  # 保持给bot看到的内容与真实用户看到的一致
                        )
    return msg, wake_up


async def get_user_name(
    event: Union[MessageEvent, GroupIncreaseNoticeEvent],
    bot: Bot,
    user_id: int,
) -> Optional[str]:
    """获取QQ用户名"""
    if (
        isinstance(event, GroupMessageEvent)
        and event.sub_type == "anonymous"
        and event.anonymous
    ):  # 匿名消息
        return f"[匿名]{event.anonymous.name}"

    if isinstance(event, (GroupMessageEvent, GroupIncreaseNoticeEvent)):
        user_info = await bot.get_group_member_info(
            group_id=event.group_id,
            user_id=user_id,
            no_cache=False,
        )
        user_name = user_info.get("nickname", None)
        user_name = user_info.get("card", None) or user_name
    else:
        user_name = event.sender.nickname if event.sender else event.get_user_id()

    return user_name



from typing import Type

from nonebot import on_command
from nonebot.adapters import Message
from nonebot.adapters.onebot.v11 import (
    MessageEvent,
)
from nonebot.log import logger
from nonebot.matcher import Matcher
from nonebot.params import CommandArg

from nonebot_plugin_dice_narrator import config
from nonebot_plugin_dice_narrator.narrator import run_narrator


def register_matcher():
    """注册消息响应器"""

    identity: Type[Matcher] = on_command(
        "检定",
        aliases={"判定", "鉴定"},
        priority=20,
        block=True,
    )

    @identity.handle()
    async def _(
        matcher: Matcher,
        event: MessageEvent,
        arg: Message = CommandArg(),
    ):
        global is_progress  # 是否产生编辑进度
        is_progress = False
        # 判断是否是禁止使用的用户
        if event.get_user_id() in config.FORBIDDEN_USERS:
            await identity.finish(
                f"您的账号({event.get_user_id()})已被禁用，请联系管理员。",
            )

        raw_cmd: str = arg.extract_plain_text()
        logger.info(f"接收到指令: {raw_cmd}")

        if raw_cmd:
            await run_narrator(question=raw_cmd, matcher=matcher)



from typing import Dict

import yaml


def export_data2yml_str(data) -> str:
    """导出字典为yml字符串"""

    return yaml.dump(data, allow_unicode=True, sort_keys=False)


def read_yml_str2data(s: str) -> Dict:
    """读取yml字符串为字典"""

    return yaml.safe_load(s)



import random
import re

from nonebot.matcher import Matcher

from nonebot_plugin_dice_narrator.config import config
from nonebot_plugin_dice_narrator.utils.common import read_yml_str2data
from nonebot_plugin_dice_narrator.utils.openai_funcs import gen_chat_response_text

SYSTEM_PROMPT = f"""
你是一个游戏DM，接下来用户会向你描述一项玩家任务，并进行一次掷骰，请给出这个任务的执行难度(1~{config.DIFFICULTY_DICE_MAX})，需要说明具体的**玩家行为描述**、四种掷骰结果对应的**事件描述**：

大成功(completed_success)：掷骰结果为{config.DIFFICULTY_DICE_MAX}
成功(success)：掷骰结果不小于任务难度
失败(failure)：掷骰结果小于任务难度
大失败(critical_failure)：掷骰结果为1

示例如下：

检定任务：
逃离触手怪的束缚

你的响应内容格式如下:

```yaml
difficulty: {int(config.DIFFICULTY_DICE_MAX*0.75)}
completed_success: 你展现出了惊人的协作和勇气。巧妙地躲过了触手怪的攻击，找到了一个隐藏的通道，成功逃离了触手怪的魔爪。你喘着粗气，但胜利的喜悦填满了你的心。
success: 你奋力抵挡住了触手怪的袭击，成功拉开了距离。虽然你没有完全逃脱，但至少还活着，可以继续寻找其他出路。
failure: 触手怪的一只触手抓住了你，你被拖入了一个黑暗的洞穴。你感到窒息和恐惧，而队友们无能为力。任务失败，你们成为了怪物的俘虏。
critical_failure: 你踩到了一块松动的地板，发出了巨大的声响。触手怪立刻察觉到了你的位置，它的触手缠绕住你，将你提起，然后狠狠地摔在地上。你昏迷了过去，不再感知周围的世界。
```

特别地，如果玩家给出的任务描述不明确，你可以适当地任意假设其发生的背景
"""

USER_PROMPT = """
检定任务:
{question}
"""

RESPONSE_STRUCT = """
检定任务: {question}
目标难度: {difficulty}/{difficulty_max}
检定结果: {dice_num} [{status}]
======
{event_content}
"""


async def run_narrator(question: str, matcher: Matcher):
    assert question, "question is empty"
    res, _ = await gen_chat_response_text(
        messages=[
            {"role": "system", "content": SYSTEM_PROMPT.strip()},
            {"role": "user", "content": USER_PROMPT.strip().format(question=question)},
        ],
    )

    # 提取并解析 yaml
    try:
        content = re.findall(r"```yaml(.*)```", res, re.S)[0]
    except:
        content = res
    try:
        res_data = read_yml_str2data(content)
    except Exception:
        res_data = read_yml_str2data(content + "```")

    # 掷骰
    dice_num = random.randint(1, config.DIFFICULTY_DICE_MAX)
    if dice_num == config.DIFFICULTY_DICE_MAX:
        res_key, status = "completed_success", "大成功"
    elif dice_num == 1:
        res_key, status = "critical_failure", "大失败"
    elif dice_num < int(res_data["difficulty"]):
        res_key, status = "failure", "失败"
    else:
        res_key, status = "success", "成功"

    print(res_data)

    await matcher.finish(RESPONSE_STRUCT.strip().format(
        question=question,
        difficulty=int(res_data["difficulty"]),
        difficulty_max=config.DIFFICULTY_DICE_MAX,
        dice_num=dice_num,
        event_content=res_data[res_key],
        status=status,
    ))



from nonebot import get_driver
from nonebot.plugin import PluginMetadata

from nonebot_plugin_dice_narrator.config import config
from nonebot_plugin_dice_narrator.matchers import register_matcher

__plugin_meta__ = PluginMetadata(
    name="nonebot-plugin-dice-narrator",
    description="",
    usage="检定 逃离触手怪的追捕",
)

global_config = get_driver().config

register_matcher()
